"""
Plot Manager for the Experiment object
"""

import numpy as np
import matplotlib.pyplot as plt
import hdfmap
from ..utils.experiment import Experiment
from .matplotlib import (
    set_plot_defaults, generate_subplots, plot_lines, plot_2d_surface, plot_3d_surface, plot_3d_lines,
    FIG_SIZE, FIG_DPI, DEFAULT_CMAP, Axes3D
)


class ExperimentPlotManager:
    """
    ExperimentPlotManager
        scan.plot = ScanPlotManager(scan)
        scan.plot() # plot default axes
        scan.plot.plot(xaxis, yaxis)  # creates figure
        scan.plot.plotline(xaxis, yaxis)  # plots line on current figure
        scan.plot.image()  # create figure and display detector image

    :param experiment: Experiment object
    """
    set_plot_defaults = set_plot_defaults
    show = plt.show

    def __init__(self, experiment: Experiment):
        self.exp = experiment

    def __call__(self, *args, **kwargs):
        return self.plot(*args, **kwargs)

    def plot(self, *scan_files: int | str, hdf_map: hdfmap.NexusMap | None = None,
             xaxis: str = 'axes', yaxis: str | list[str] = 'signal', axes: plt.Axes | None = None, **kwargs) -> plt.Axes:
        """
        Create matplotlib figure with a line plot from a or several scans

            axes = exp.plot.plot(file1, file2, xaxis='eta', yaxis='roi2_sum')

        :param scan_files: scan number or filename (multiple allowed)
        :param hdf_map: hdfmap object or None
        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis, also accepts list of names for multiple lines
        :param axes: matplotlib.axes subplot, or None to create a figure
        :param kwargs: given directly to plt.plot(..., **kwars)
        :return: axes object
        """
        scans = self.exp.scans(*scan_files, hdf_map=hdf_map)
        if axes is None:
            fig, axes = plt.subplots()
        for scan in scans:
            scan.plot.plot(xaxis, yaxis, axes=axes, **kwargs)
        if len(scan_files) > 1:
            axes.legend(scan_files)
        axes.set_title(self.exp.generate_scans_title(*scan_files))
        return axes

    def image(self,  scan_file: int | str = -1, hdf_map: hdfmap.NexusMap | None = None,
              index: int | tuple | slice | None = None, xaxis: str = 'axes',
              axes: plt.Axes | None = None, clim: tuple[float, float] | None = None,
              cmap: str = DEFAULT_CMAP, colorbar: bool = False, **kwargs) -> plt.Axes | None:
        """
        Plot image in matplotlib figure (if available)
        :param scan_file: scan number or filename
        :param hdf_map: hdfmap object or None
        :param index: int, detector image index, 0-length of scan, if None, use centre index
        :param xaxis: name or address of xaxis dataset
        :param axes: matplotlib axes to plot on (None to create figure)
        :param clim: [min, max] colormap cut-offs (None for auto)
        :param cmap: str colormap name (None for auto)
        :param colorbar: False/ True add colorbar to plot
        :param kwargs: additional arguments for plot_detector_image
        :return: axes object or None if no image
        """
        scan = self.exp.scans(scan_file, hdf_map=hdf_map)[0]
        if scan.map.image_data:
            return scan.plot.image(index, xaxis, axes, clim, cmap, colorbar, **kwargs)
        return None

    def detail(self, scan_file: int | str = -1, hdf_map: hdfmap.NexusMap | None = None,
               xaxis: str = 'axes', yaxis: str | list[str] = 'signal',
               index: int | tuple | slice | None = None, clim: tuple[float, float] | None = None,
               cmap: str = DEFAULT_CMAP, **kwargs) -> plt.Figure:
        """
        Create matplotlib figure with plot of the scan and detector image
        :param scan_file: scan number or filename
        :param hdf_map: hdfmap object or None
        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis, also accepts list of names for multiple plots
        :param index: int, detector image index, 0-length of scan, if None, use centre index
        :param clim: [min, max] colormap cut-offs (None for auto)
        :param cmap: str colormap name (None for auto)
        :param kwargs: given directly to plt.plot(..., *args, **kwars)
        :return: figure object
        """
        scan = self.exp.scans(scan_file, hdf_map=hdf_map)[0]

        # Create figure
        fig, ((lt, rt), (lb, rb)) = plt.subplots(2, 2, figsize=[FIG_SIZE[0] * 1.2, FIG_SIZE[1] * 1.2], dpi=FIG_DPI)
        fig.subplots_adjust(hspace=0.35, left=0.1, right=0.95)
        fig.suptitle(scan.title())

        # Top left - line plot
        scan.plot(xaxis, yaxis, axes=lt, **kwargs)
        lt.set_title(None)

        # Top right - image plot
        scan.plot.image(index, xaxis, cmap=cmap, clim=clim, axes=rt)
        rt.set_title(None)

        # Bottom-Left - details
        details = self.exp.scan_str(scan_file, hdf_map=hdf_map)
        lb.text(-0.2, 1, details, ha='left', va='top', multialignment="left", fontsize=12, wrap=True)
        lb.set_axis_off()

        # Bottom-Right - fit results
        rb.set_axis_off()
        if 'fit' in yaxis:
            fit_report = str(scan.fit)
            rb.text(-0.2, 1, fit_report, ha='left', va='top', multialignment="left", fontsize=12, wrap=True)
        return fig

    def multi_lines(self, *scan_files: int | str, hdf_map: hdfmap.NexusMap | None = None,
                    xaxis: str = 'axes', yaxis: str = 'signal', value: str | None = None,
                    axes: plt.Axes | None = None, **kwargs) -> plt.Axes:
        """
        Create matplotlib figure with a line plot from a or several scans

            axes = exp.plot.multi_lines(*range(-10, 0), xaxis='eta', yaxis='roi2_sum', value='Tsample')

        :param scan_files: scan number or filename (multiple allowed)
        :param hdf_map: hdfmap object or None
        :param xaxis: str name or path of array to plot on x axis
        :param yaxis: str name or path of array to plot on y axis
        :param value: str name or path of float value to distinguish lines
        :param axes: matplotlib.axes subplot, or None to create a figure
        :param kwargs: given directly to plt.plot(..., **kwars)
        :return: axes object
        """
        scans = self.exp.scans(*scan_files, hdf_map=hdf_map)
        hdf_map = scans[0].map

        plot_data = []
        for scan in scans:
            with hdfmap.load_hdf(scan.filename) as hdf:
                xdata = hdf_map.eval(hdf, xaxis)
                ydata = hdf_map.eval(hdf, yaxis)
                value_data = np.mean(hdf_map.eval(hdf, value)) if value else None
                plot_data.append((value_data, xdata, ydata))

        if axes is None:
            fig, axes = plt.subplots()
        lines, sm = plot_lines(axes, *plot_data, **kwargs)

        x_lab, y_lab, value_label = hdf_map.generate_ids(xaxis, yaxis, value, modify_missing=False)
        axes.set_xlabel(x_lab)
        axes.set_ylabel(y_lab)
        axes.set_title(self.exp.generate_scans_title(*scan_files, hdf_map=hdf_map))

        plt.colorbar(sm, ax=axes, label=value_label)
        return axes

    def multi_plot(self, *scan_files: int | str, hdf_map: hdfmap.NexusMap | None = None,
                   xaxis: str = 'axes', yaxis: str | list[str] = 'signal', value: str | None = None,
                   subplots: tuple[int, int] = (4, 4), **kwargs) ->  list[tuple[plt.Figure, plt.Axes]]:
        """
        Create matplotlib figure with a line plot from a or several scans

            axes = exp.plot.multi_plot(*range(-10, 0), xaxis='eta', yaxis='roi2_sum', value='Tsample')

        :param scan_files: scan number or filename (multiple allowed)
        :param hdf_map: hdfmap object or None
        :param xaxis: str name or path of array to plot on x axis
        :param yaxis: str name or path of array to plot on y axis, or a list of names for multiple plots per axis
        :param value: str name or path of float value to distinguish lines
        :param subplots: [int, int] number of subplots per figure [vertical, horizontal]
        :param kwargs: given directly to plt.plot(..., **kwars)
        :return: list of figures
        """
        scans = self.exp.scans(*scan_files, hdf_map=hdf_map)
        ttl = self.exp.generate_scans_title(*scan_files)
        fig_ax = generate_subplots(len(scan_files), subplots=subplots, suptitle=ttl)

        for scan, (fig, ax) in zip(scans, fig_ax):
            scan.plot.plot(xaxis=xaxis, yaxis=yaxis, axes=ax, **kwargs)
            val = scan.format("\n%s = {%s}" % (value, value)) if value is not None else ""
            ttl_exp = scan.title() + val
            ax.set_title(ttl_exp)
        return fig_ax

    def surface_2d(self, *scan_files: int | str, hdf_map: hdfmap.NexusMap | None = None,
                   xaxis: str = 'axes', signal: str = 'signal', values: str | None = None,
                   axes: plt.Axes | None = None, clim: tuple[float, float] | None = None,
                   axlim: str = 'image', **kwargs) -> plt.Axes:
        """
        Create matplotlib figure with a 2D image

            axes = exp.plot.surface_2d(*range(-10, 0), xaxis='eta', signal='roi2_sum', values='Tsample')

        :param scan_files: scan number or filename (multiple allowed)
        :param hdf_map: hdfmap object or None
        :param xaxis: str name or path of array to plot on x axis
        :param signal: str name or path of array to plot on z axis
        :param values: str name or path of float value to distinguish different scans
        :param axes: matplotlib.axes subplot, or None to create a figure
        :param clim: None or [min, max] values for color cutoff from plt.clim
        :param axlim: axis limits from plt.axis
        :param kwargs: given directly to plt.pcolormesh(..., **kwargs)
        :return: axes object
        """
        first_file = self.exp.get_scan_filename(scan_files[0])
        hdf_map = hdfmap.create_nexus_map(first_file) if hdf_map is None else hdf_map
        x, y, z = self.exp.generate_mesh(*scan_files, hdf_map=hdf_map,
                                         axes=xaxis, signal=signal, values=values)

        if axes is None:
            fig, axes = plt.subplots()
        surf = plot_2d_surface(axes=axes, image=z, xdata=x, ydata=y, clim=clim, axlim=axlim, **kwargs)
        axes.set_title(self.exp.generate_scans_title(*scan_files))

        x_lab, y_lab, value_label = hdf_map.generate_ids(xaxis, signal, values, modify_missing=False)
        axes.set_xlabel(x_lab)
        axes.set_ylabel(value_label)
        axes.figure.colorbar(surf, ax=axes, label=signal)
        return axes

    def lines_3d(self, *scan_files: int | str, hdf_map: hdfmap.NexusMap | None = None,
                 xaxis: str = 'axes', signal: str = 'signal', values: str | None = None,
                 axes: Axes3D | None = None, legend: bool = False, **kwargs) -> Axes3D:
        """
        Create matplotlib figure with a 2D image

            axes = exp.plot.surface_2d(*range(-10, 0), xaxis='eta', yaxis='roi2_sum', value='Tsample')

        :param scan_files: scan number or filename (multiple allowed)
        :param hdf_map: hdfmap object or None
        :param xaxis: str name or path of array to plot on x axis
        :param signal: str name or path of array to plot on z axis
        :param values: str name or path of float value to distinguish different scans
        :param axes: matplotlib.axes subplot, or None to create a figure
        :param legend: if True, adds a legend
        :param kwargs: given directly to plt.plot(..., **kwargs)
        :return: axes object
        """
        first_file = self.exp.get_scan_filename(scan_files[0])
        hdf_map = hdfmap.create_nexus_map(first_file) if hdf_map is None else hdf_map
        scans = self.exp.scans(*scan_files, hdf_map=hdf_map)
        data_fields = [signal, xaxis] + ([values] if values is not None else [])
        data = self.exp.join_scan_data(*scan_files, hdf_map=hdf_map, data_fields=data_fields)
        labels = [scan.label() for scan in scans]

        if axes is None:
            fig = plt.figure()
            axes = fig.add_subplot(111, projection='3d')
        plot_3d_lines(axes, data[signal], data[xaxis], data.get(values, None), labels=labels, **kwargs)
        axes.set_title(self.exp.generate_scans_title(*scan_files))

        x_lab, signal_label, value_label = hdf_map.generate_ids(xaxis, signal, values, modify_missing=False)
        axes.set_xlabel(x_lab)
        axes.set_ylabel(value_label)
        axes.set_zlabel(signal_label)
        if legend:
            axes.legend()
        return axes

    def surface_3d(self, *scan_files: int | str, hdf_map: hdfmap.NexusMap | None = None,
                   xaxis: str = 'axes', signal: str = 'signal', values: str | None = None,
                   axes: Axes3D | None = None, clim: tuple[float, float] | None = None,
                   axlim: str = 'image', **kwargs) -> Axes3D:
        """
        Create matplotlib figure with a 3D image

            axes = exp.plot.surface_3d(*range(-10, 0), xaxis='eta', yaxis='roi2_sum', value='Tsample')

        :param scan_files: scan number or filename (multiple allowed)
        :param hdf_map: hdfmap object or None
        :param xaxis: str name or path of array to plot on x axis
        :param signal: str name or path of array to plot on z axis
        :param values: str name or path of float value to distinguish different scans
        :param axes: matplotlib.axes3D subplot, or None to create a figure
        :param clim: None or [min, max] values for color cutoff from plt.clim
        :param axlim: axis limits from plt.axis
        :param kwargs: given directly to plt.pcolormesh(..., **kwargs)
        :return: axes object
        """
        first_file = self.exp.get_scan_filename(scan_files[0])
        hdf_map = hdfmap.create_nexus_map(first_file) if hdf_map is None else hdf_map
        x, y, z = self.exp.generate_mesh(*scan_files, hdf_map=hdf_map,
                                         axes=xaxis, signal=signal, values=values)

        if axes is None:
            fig = plt.figure()
            axes = fig.add_subplot(111, projection='3d')
        plot_3d_surface(axes=axes, image=z, xdata=x, ydata=y, clim=clim, axlim=axlim, **kwargs)
        axes.set_title(self.exp.generate_scans_title(*scan_files))

        x_lab, signal_label, value_label = hdf_map.generate_ids(xaxis, signal, values, modify_missing=False)
        axes.set_xlabel(x_lab)
        axes.set_ylabel(value_label)
        axes.set_zlabel(signal_label)
        return axes

    def metadata(self, *scan_files: int | str, values: str | list[str], hdf_map: hdfmap.NexusMap | None = None,
                 axes: plt.Axes | None = None) -> plt.Axes:
        """
        Create matplotlib figure with plot of metadata vs scan number

        :param scan_files: scan number or filename (multiple allowed)
        :param values: field name or path of float value to distinguish different scans
        :param hdf_map: hdfmap object or None
        :param axes: matplotlib.axes subplot, or None to create a figure
        :return: axes object
        """
        scans = self.exp.scans(*scan_files, hdf_map=hdf_map)
        ttl = self.exp.generate_scans_title(*scan_files)

        values = [values] if isinstance(values, str) else values
        data = {
            name: np.array([scan.values(name) for scan in scans])
            for name in values
        }

        x_data = [scan.scan_number() for scan in scans]
        if axes is None:
            fig, axes = plt.subplots()
        # TODO: handle 2D grid scans
        for name in data:
            axes.plot(x_data, data[name], '-o', label=name)
        axes.set_xlabel('Scan number')
        axes.legend()
        axes.set_title(ttl)
        return axes



"""
Spectra object

=== DATA MODEL ===
spectra = Spectra(energy, signal, mode, process)
spectra2 = spectra + 2  # adds 2 to signal
spectra3 = spectra.norm_to_peak()
average_spectra = spectra2 + spectra3  # averages spectra signals at interpolated energies
subtracted_spectra = spectra2 - spectra3  # subtracts signals at interpolated energies
"""
from __future__ import annotations

from inspect import signature
from typing_extensions import Self
import numpy as np
from matplotlib.axes import Axes
import matplotlib.pyplot as plt
import h5py

from mmg_toolbox.nexus import nexus_writer as nw
from mmg_toolbox.xas import spectra_analysis as spa


# Build database of spectra functions from spectra_analysis.py
# these will be called by Spectra.remove_background
def get_func_doc(fn):
    sig = signature(fn)
    doc = next(iter(fn.__doc__.splitlines()), 'background function')
    args = ". args: " + ', '.join(str(par) for name, par in sig.parameters.items() if name not in ['energy', 'signal'])
    return doc + args

BACKGROUND_FUNCTIONS = {
    # description: (en, sig, *args, **kwargs) -> bkg, norm, fit
    'flat': spa.subtract_flat_background,  # ev_from_start
    'norm': spa.normalise_background,  # ev_from_start
    'linear': spa.fit_linear_background,  # ev_from_start
    'curve': spa.fit_curve_background,  # ev_from_start
    'exp': spa.fit_exp_background,  # ev_from_start, ev_from_end
    'step': spa.fit_step_background,  # ev_from_start
    'double_edge_step': spa.fit_double_edge_step_background,  # l3_energy, l2_energy, peak_width_ev
    'poly_edges': spa.fit_spectra_background, # *step_energies, peak_width_ev
    'exp_edges': spa.fit_spectra_exp_background, # *step_energies, peak_width_ev
}
BACKGROUND_DOCSTRINGS = {
    # name: next(iter(fn.__doc__.splitlines()), 'background function') for name, fn in BACKGROUND_FUNCTIONS.items()
    # name: f"{fn.__name__}{str(inspect.signature(fn))}" for name, fn in BACKGROUND_FUNCTIONS.items()
    name: get_func_doc(fn) for name, fn in BACKGROUND_FUNCTIONS.items()
}


class Spectra:
    """
    An energy spectra, containing:

    :param energy: n-length array
    :param signal: n-length array
    :param background: n-length array (*or None)
    :param parents: list of Spectra or Scan objects (*or None)
    :param mode: str name of the spectra for plots
    :param process_label: str label of the process
    :param process: str describing a process done to the spectra
    """
    def __init__(self, energy: np.ndarray, signal: np.ndarray,
                 background: np.ndarray | None = None,
                 parents: list[Self] | None = None,
                 mode: str = '', process_label: str = 'raw', process: str = ''):
        if parents is None:
            parents = []
        self.parents = parents
        self.energy = energy
        self.signal = signal
        if energy.shape != signal.shape:
            raise Exception(f"the shape of energy[{energy.shape}] and signal[{signal.shape}] must match")
        self.background = background
        if background is not None and background.shape != energy.shape:
            raise Exception(f"the shape of energy[{energy.shape}] and background[{background.shape}] must match")
        self.process_label = process_label
        self.process = process
        self.mode = mode

    """SPECTRA PROPERTIES"""

    def __repr__(self):
        return f"Spectra('{self.mode}', energy=array{self.energy.shape}, signal=array{self.signal.shape}, process_label='{self.process_label}')"

    def edge_label(self, edges=spa.SEARCH_EDGES):
        return spa.energy_range_edge_label(self.energy.min(), self.energy.max(), search_edges=edges)

    def edges(self, search_edges=spa.SEARCH_EDGES):
        return spa.xray_edges_in_range(self.energy.min(), self.energy.max(), search_edges=search_edges)

    def signal_at_energy(self, energy1: float, energy2: float | None = None) -> float:
        """Return averaged signal between energy values"""
        idx1 = np.argmin(np.abs(self.energy - energy1))
        if energy2 is None:
            idx2 = idx1 + 1
        else:
            idx2 = np.argmin(np.abs(self.energy - energy2))
        return float(np.mean(self.signal[idx1:idx2]))

    def signal_peak(self) -> float:
        return np.max(abs(self.signal))

    def signal_jump(self, ev_from_start=5., ev_from_end=None) -> float:
        return spa.signal_jump(self.energy, self.signal, ev_from_start, ev_from_end)

    """SPECTRA OPERATIONS - RETURNS PROCESSED SPECTRA"""

    def __add__(self, other) -> Spectra:
        if issubclass(type(other), Spectra):
            return SpectraAverage(self, other)
        return Spectra(self.energy, self.signal + other, self.background, mode=self.mode,
                       parents=[self], process_label='add_value', process=f"{self.mode} + {other}")

    def __mul__(self, other) -> Spectra:
        if issubclass(type(other), Spectra):
            raise TypeError("Cannot multiply Spectra")
        return Spectra(self.energy, self.signal * other, self.background, mode=self.mode,
                       parents=[self], process_label='multiply', process=f'{self.mode} * {other}')

    def __sub__(self, other) -> Spectra:
        if other in self.parents:
            # remove other, recalculate average
            return SpectraAverage(*(parent for parent in self.parents if parent != other))
        if issubclass(type(other), Spectra):
            # subtract new spectra from this spectra
            return SpectraSubtraction(self, other)
        return Spectra(self.energy, self.signal - other, self.background, mode=self.mode,
                       parents=[self], process_label='subtract_value', process=f'{self.mode}-{other}')

    def divide_by_signal_at_energy(self, energy1: float, energy2: float | None = None) -> Spectra:
        """Divide spectra by signal"""
        value = self.signal_at_energy(energy1, energy2)
        sig = self.signal / value
        bkg = self.background / value if self.background is not None else None
        proc_label = "divide_by_signal_at_energy"
        process = f"normalise  to signal at energy between {energy1:.0f} and {energy2} eV\n"
        process += f"energy1 = {energy1}\n"
        process += f"energy2 = {energy2}\n"
        process += f"Spectra.signal_at_energy(energy1, energy2) = {value}\n"
        process += f"Spectra.signal = signal / {value:.3f}"
        return Spectra(self.energy, sig, parents=[self], background=bkg,
                       process_label=proc_label, process=process, mode=self.mode)

    def divide_by_preedge(self, ev_from_start: float = 5) -> Spectra:
        """Divide by average of raw_signals at start"""
        value = spa.preedge_signal(self.energy, self.signal, ev_from_start)
        sig = self.signal / value
        bkg = self.background / value if self.background is not None else None
        proc_label = 'divide_by_preedge'
        process = f"normalise signal to signal in the pre-edge region in first {ev_from_start} eV\n"
        process += f"mean(Spectra.signal[:{ev_from_start}]) = {value}\n"
        process += f"Spectra.signal = signal / {value:.3f}"
        return Spectra(self.energy, sig, parents=[self], background=bkg,
                       process_label=proc_label, process=process, mode=self.mode)

    def divide_by_postedge(self, ev_from_end: float = 5) -> Spectra:
        """Divide by average of raw_signals at start"""
        value = spa.postedge_signal(self.energy, self.signal, ev_from_end)
        sig = self.signal / value
        bkg = self.background / value if self.background is not None else None
        proc_label = 'divide_by_postedge'
        process = f"normalise signal to signal in the post-edge region in last {ev_from_end} eV\n"
        process += f"mean(Spectra.signal[:{ev_from_end} eV]) = {value}\n"
        process += f"Spectra.signal = signal / {value:.3f}"
        return Spectra(self.energy, sig, parents=[self], background=bkg,
                       process_label=proc_label, process=process, mode=self.mode)

    def norm_to_peak(self) -> Spectra:
        peak = self.signal_peak()
        sig = self.signal / peak
        bkg = self.background / peak if self.background is not None else None
        proc_label = 'norm_to_peak'
        process = f"normalise signal to the maximum peak height\n"
        process += f"max(Spectra.signal) = {peak}\n"
        process += f"Spectra.signal = signal / {peak:.3f}"
        return Spectra(self.energy, sig, parents=[self], background=bkg,
                       process_label=proc_label, process=process, mode=self.mode)

    def norm_to_jump(self, ev_from_start=5., ev_from_end=None) -> Spectra:
        jump = abs(self.signal_jump(ev_from_start, ev_from_end))
        sig = self.signal / jump
        bkg = self.background / jump if self.background is not None else None
        proc_label = 'norm_to_jump'
        process = f"normalise signal to the jump in signal between start and end of spectra\n"
        process += f"ev_from_start = {ev_from_start}\n"
        process += f"ev_from_end = {ev_from_end}\n"
        process += f"jump(Spectra.signal) = {jump}\n"
        process += f"Spectra.signal = signal / {jump:.3f}"
        return Spectra(self.energy, sig, parents=[self], background=bkg,
                       process_label=proc_label, process=process, mode=self.mode)

    def remove_background(self, name='flat', *args, **kwargs) -> Spectra:
        """
        Return new Spectra object with background removed
        """
        bkg_fun = BACKGROUND_FUNCTIONS[name]
        bkg_doc = BACKGROUND_DOCSTRINGS[name]
        bkg, norm, fit = bkg_fun(self.energy, self.signal, *args, **kwargs)
        sig = (self.signal - bkg) / norm
        proc_label = f"{name}"
        process = f"Background removal '{name}', using function: \n    {bkg_doc}\n"
        process += f"args: {str(args)}\n"
        process += f"kwargs: {str(kwargs)}\n"
        process += f"\nFit Report:\n{fit.fit_report() if fit is not None else 'None'}\n"
        process += f"\nResults:\n  <bkg> = {np.mean(bkg)}\n  norm = {norm}\nsignal = (signal - bkg) / norm\n"
        return Spectra(self.energy, sig, parents=[self], background=bkg,
                       process_label=proc_label, process=process, mode=self.mode)
    remove_background.__doc__ += (
            "available functions:\n" +
            '\n'.join(f"'{name}': {doc}" for name, doc in BACKGROUND_DOCSTRINGS.items())
    )

    def auto_edge_background(self, peak_width_ev: float = 5.) -> Spectra:
        """
        Remove generic xray absorption background from spectra
        """
        edges = self.edges()
        edge_energies = [energy for lbl, energy in edges]
        bkg, jump, fit = spa.fit_spectra_background(self.energy, self.signal, *edge_energies, peak_width_ev=peak_width_ev)
        sig = (self.signal - bkg) / jump
        proc_label = f"Auto_edge_background"
        process = f"Background removal 'poly_edges', using function: \n    {BACKGROUND_DOCSTRINGS['poly_edges']}\n"
        process += f"peak_width_ev = {peak_width_ev}\n"
        edge_str = '\n'.join(f"  {lbl}: {energy}" for lbl, energy in edges)
        process += f"Edges:\n {edge_str}"
        process += f"\nFit Report:\n{fit.fit_report() if fit is not None else 'None'}\n"
        process += f"\nResults:\n  <bkg> = {np.mean(bkg)}\n  norm = {jump}\nsignal = (signal - bkg) / norm\n"
        return Spectra(self.energy, sig, parents=[self], background=bkg,
                       process_label=proc_label, process=process, mode=self.mode)

    """SPECTRA NEXUS OUTPUT"""

    def create_nxnote(self, parent: h5py.Group, name: str, sequence_index: int | None = None) -> h5py.Group:
        note = nw.add_nxnote(
            root=parent,
            name=name,
            description=f"{self.mode} {self.process_label}",
            data=self.process,
            sequence_index=sequence_index
        )
        return note

    def create_nxdata(self, parent: h5py.Group, name: str, default: bool = False) -> h5py.Group:
        """NXxas NXdata entry, inlcuding energy, absorped beam and mode"""
        data = nw.add_nxdata(parent, name, axes=['energy'], signal='absorbed_beam')
        nw.add_nxfield(data, 'mode', self.mode)
        nw.add_nxfield(data, 'energy', self.energy, units='eV')
        nw.add_nxfield(data, 'absorbed_beam', self.signal, units='')
        if default:
            parent.attrs['default'] = name
        return data

    """SPECTRA PLOT FUNCTIONS"""

    def plot(self, ax: Axes | None = None, *args, **kwargs) -> list[plt.Line2D]:
        if ax is None:
            ax = plt.gca()
        if 'label' not in kwargs:
            kwargs['label'] = f"{self.mode} {self.process_label}"
        return ax.plot(self.energy, self.signal, *args, **kwargs)

    def plot_bkg(self, ax: Axes | None = None, *args, **kwargs) -> list[plt.Line2D]:
        if self.background is None:
            return []
        if ax is None:
            ax = plt.gca()
        if 'label' not in kwargs:
            kwargs['label'] = f"{self.mode} {self.process_label} bkg"
        return ax.plot(self.energy, self.background, *args, **kwargs)

    def plot_parents(self, ax: Axes | None = None, *args, **kwargs) -> list[plt.Line2D]:
        """Plot all parents on the current axes"""
        if ax is None:
            ax = plt.gca()
        pl = []
        label = kwargs.get('label', '')
        for parent in self.parents:
            if issubclass(type(parent), Spectra):
                kwargs['label'] = parent.mode + label
                pl += ax.plot(parent.energy, parent.signal, *args, **kwargs)
        return pl

    def create_figure(self) -> plt.Figure:
        """Create figure with spectra plot"""
        fig, ax1 = plt.subplots(1, 1)
        self.plot_parents(ax1)
        self.plot_bkg(ax1)
        self.plot(ax1)

        ax1.set_xlabel('E [eV]')
        ax1.set_ylabel('signal')
        ax1.legend()
        return fig


class SpectraSubtraction(Spectra):
    """Difference between two spectra"""

    def __init__(self, spectra1: Spectra, spectra2: Spectra):
        if spectra1.mode != spectra2.mode:
            raise ValueError('Spectra1 and spectra2 must have same mode')
        parents = [spectra1, spectra2]
        av_energy = spa.average_energy_scans(spectra1.energy, spectra2.energy)
        signal1 = spa.average_energy_spectra(av_energy, (spectra1.energy, spectra1.signal))
        signal2 = spa.average_energy_spectra(av_energy, (spectra2.energy, spectra2.signal))
        if all(s.background is None for s in parents):
            av_bkg = None
        else:
            bkg_spectra = ((s.energy, s.background) for s in parents if s.background is not None)
            av_bkg = spa.average_energy_spectra(av_energy, *bkg_spectra)
        difference = signal1 - signal2
        label = next((s.mode for s in parents))
        process_label = 'subtraction'
        process = "Subtraction of spectra S1 - S2:\n" + "\n".join(
            f'  S{n + 1}: ' + repr(s) for n, s in enumerate(parents))
        super().__init__(av_energy, difference, av_bkg, parents,
                         process_label=process_label, mode=label, process=process)

    def __repr__(self):
        return (
            f"SpectraSubtraction('{self.mode}', energy=array{self.energy.shape}, signal=array{self.signal.shape}," +
            f"process_label='{self.process_label}')"
        )

    def average_subtracted_spectra(self):
        spectra1, spectra2 = self.parents
        average = spectra1 + spectra2
        return average

    def calculate_sum_rules(self, n_holes: float) -> tuple[float, float]:
        difference = self.signal
        average = self.average_subtracted_spectra().signal
        orb = spa.orbital_angular_momentum(self.energy, average, difference, n_holes)
        spin = spa.spin_angular_momentum(self.energy, average, difference, n_holes)
        return orb, spin

    def sum_rules_report(self, n_holes: float, element: str = '') -> str:
        orb, spin = self.calculate_sum_rules(n_holes)
        report = f"{element} n_holes = {n_holes}\nL = {orb:.3f} μB\nS = {spin:.3f} μB"
        return report

    def create_sum_rules_nxnote(self, n_holes: float, parent: h5py.Group,
                                name: str, sequence_index: int | None = None, element: str = '') -> h5py.Group:
        note = nw.add_nxnote(
            root=parent,
            name=name,
            description=f"{self.mode} {self.process_label} Sum Rules",
            data=self.sum_rules_report(n_holes, element),
            sequence_index=sequence_index
        )
        return note


class SpectraAverage(Spectra):
    """Averaged Spectra"""
    def __init__(self, *spectra: Spectra):
        mode = next(iter(s.mode for s in spectra if s.mode is not None), '')
        process_label = 'average'
        spectra = [s for s in spectra if s.mode == mode]
        # avoid average of average by using parents of previously averaged spectra
        parents = []
        for s in spectra:
            if len(s.parents) > 1:
                parents += s.parents
            else:
                parents.append(s)
        av_energy = spa.average_energy_scans(*(s.energy for s in parents))
        av_signal = spa.average_energy_spectra(av_energy, *((s.energy, s.signal) for s in parents))
        if all(s.background is None for s in parents):
            av_bkg = None
        else:
            bkg_spectra = ((s.energy, s.background) for s in parents if s.background is not None)
            av_bkg = spa.average_energy_spectra(av_energy, *bkg_spectra)
        label = next((s.mode for s in parents))
        process = "Average of spectra:\n" + "\n".join('  ' + repr(s) for s in parents)
        super().__init__(av_energy, av_signal, av_bkg, parents, mode=label,
                         process_label=process_label, process=process)

    def __repr__(self):
        return (
            f"SpectraAverage('{self.mode}', energy=array{self.energy.shape}, signal=array{self.signal.shape}," +
            f"process_label='{self.process_label}')"
        )




import os

from mmg_toolbox.beamline_metadata.config import beamline_config
import hdfmap
from hdfmap import load_hdf, create_nexus_map

from mmg_toolbox.nexus.nexus_scan import NexusDataHolder, NexusScan
from mmg_toolbox.utils.file_functions import get_scan_number, replace_scan_number


def read_nexus_file(filename: str, flatten_scannables: bool = True, beamline: str | None = None) -> NexusDataHolder:
    """
    Read Nexus file as DataHolder
    """
    config = beamline_config(beamline)
    return NexusDataHolder(filename, flatten_scannables=flatten_scannables, config=config)


def read_nexus_files(*filenames: str, beamline: str | None = None) -> list[NexusScan]:
    """
    Read Nexus files as NexusScan
    """
    hdf_map = create_nexus_map(filenames[0])
    config = beamline_config(beamline)
    return [NexusScan(f, hdf_map, config=config) for f in filenames]


def find_matching_scans(filename: str, match_field: str = 'scan_command',
                        search_scans_before: int = 10, search_scans_after: int | None = None) -> list[str]:
    """
    Find scans with scan numbers close to the current file with matching scan command

    :param filename: nexus file to start at (must include scan number in filename)
    :param match_field: nexus field to compare between scan files
    :param search_scans_before: number of scans before current scan to look for
    :param search_scans_after: number of scans after current scan to look for (None==before)
    :returns: list of scan files that exist and have matching field values
    """
    nexus_map = create_nexus_map(filename)
    field_value = nexus_map.eval(nexus_map.load_hdf(), match_field)
    scanno = get_scan_number(filename)
    if search_scans_after is None:
        search_scans_after = search_scans_before
    matching_files = []
    for scn in range(scanno - search_scans_before, scanno + search_scans_after):
        new_filename = replace_scan_number(filename, scn)
        if os.path.isfile(new_filename):
            new_field_value = nexus_map.eval(load_hdf(new_filename), match_field)
            if field_value == new_field_value:
                matching_files.append(new_filename)
    return matching_files


def find_similar_scans(filename: str, *files: str, metadata: list[str | tuple[str, float]]) -> list[str]:
    """
    Find scan files with similar metadata

        metadata = [
            'name',  # match scans with file2['name'] == file1['name']
            ('name', tol)  # match scans with abs(file2['name'] - file1['name']) < tol
        ]
        match_files = find_similar_scans(*filenames, metadata=metadata)

    :param filename: nexus scan file to compare others with
    :praam files: list of additional nexus files
    :param metadata: list of nexus metadata fields to compare
    :returns: list of scan files that match all requirements
    """
    nexus_map = create_nexus_map(filename)
    names = [n if isinstance(n, str) else n[0] for n in metadata]
    with hdfmap.load_hdf(filename) as hdf:
        initial_parameters = [
            nexus_map.eval(hdf, name)
            for name in names
        ]
    similar_files = []
    for file in files:
        if file == filename:
            continue
        with hdfmap.load_hdf(file) as hdf:
            new_parameters = [
                nexus_map.eval(hdf, name)
                for name in names
            ]
        match = all(
            ini == new if isinstance(par, str) else abs(ini - new) < par[1]
            for par, ini, new in zip(metadata, initial_parameters, new_parameters)
        )
        if match:
            similar_files.append(file)
    return similar_files


def find_scans(filename: str, *files: str, hdf_map: hdfmap.NexusMap | None = None, first_only: bool = False,
               **matches: str | float | tuple[float, float]) -> list[str]:
    """
    Find scans files with matching parameters

        matches = {
            'name1': 'scan', # matches if 'scan' in file['name1']
            'name2': value, # matches if file['name2'] ~= value
            'name3': (value, tol), # matches if abs(file['name3'] - value) < tol
        }
        match_files = find_scans(*filenames, **matches)

    :param filename: nexus filename (used to create hdfmap)
    :param files: list of additional nexus files
    :param hdf_map: if given, uses this hdfmap rather than generating one.
    :param first_only: if true, returns on the first result
    :param matches: keyword arguments for matching parameters
    :returns: list of scan files that match all requirements
    """
    nexus_map = hdf_map or create_nexus_map(filename)
    matching_files = []
    for file in (filename,) + files:
        with hdfmap.load_hdf(file) as hdf:
            all_ok = True
            for name, match_value in matches.items():
                file_value = nexus_map.eval(hdf, name)
                if isinstance(match_value, str):
                    chk = match_value in file_value
                elif isinstance(match_value, (float, int)):
                    chk = abs(match_value - file_value) < 0.01
                else:
                    chk = abs(match_value[0] - file_value) < match_value[1]
                if not chk:
                    all_ok = False
                    break
        if all_ok:
            matching_files.append(file)
            if first_only:
                break
    return matching_files




"""
Specific beamline metadata
"""

META_LABEL = "{(cmd|user_command|scan_command)}"

META_STRING = """
{filename}
{filepath}
{start_time}
cmd = {(cmd|user_command|scan_command)}
axes = {_axes}
signal = {_signal}
detector = {_IMAGE.split('/')[-2] if '/' in _IMAGE else 'none'}
shape = {axes.shape}
"""

I06_1_META_STRING = META_STRING + """
endstation: {end_station}
sample = {sample_name}
energy = {mean((energyh|incident_energy)):.0f} eV
pol = {polarisation}
field = {field_x?(0)**2 + field_y?(0)**2 + field_z?(0)**2:.2f} T
temp = {(lakeshore336_cryostat|itc3_device_sensor_temp?(300)):.2f} K
sample y pos = {scm_y?(0):.2f} 
pitch = {m7_pitch?(0):.2f}
"""

I10_1_META_STRING = META_STRING + """
endstation: {endstation}
sample = {sample_name}
energy = {mean(energyh?(0)):.0f} eV
pol = {polarisation}
field = {(magnet_field|ips_demand_field?(0)):.2f} T
temp = {(lakeshore336_cryostat|itc3_device_sensor_temp?(300)):.2f} K
sample y pos = {(em_y|hfm_y?(0)):.2f} 
pitch = {(em_pitch|hfm_pitch?(0)):.2f}
"""

I16_META_STRING = META_STRING + """
sample = {sample_name}
energy = {mean(incident_energy):.0f} keV
pol = {mean(stokes):.2f} analyser: {analyser_name}({analyser_order}) {pa_detector_name}
temp = {Tsample?(300):.2f} K

hkl = ({mean(diffractometer_sample_h):.2g},{mean(diffractometer_sample_k):.2g},{mean(diffractometer_sample_l):.2g})
psi = {mean(diffractometer_sample_psi?(nan)):.2f} Deg, azir=({diffractometer_sample_azih?(0)},{diffractometer_sample_azik?(0)},{diffractometer_sample_azil?(0)})
sx = {mean(sx):6.2f}, sy = {mean(sy):6.2f}, sz = {mean(sz):6.2f} mm
eta = {mean(diffractometer_sample_eta):6.2f}, chi = {mean(diffractometer_sample_chi):6.2f}, phi = {mean(diffractometer_sample_phi):6.2f}, mu = {mean(diffractometer_sample_mu):6.2f}  Deg
delta = {mean(diffractometer_sample_delta):6.2f}, gamma = {mean(diffractometer_sample_gam):6.2f}  Deg

Atten = {Atten} ({100 * Transmission: .3g} %)
ss = [{mean(s5xgap):.3g}, {mean(s5ygap):.3g}]
ds = [{mean(s7xgap):.3g}, {mean(s7ygap):.3g}]
"""

BEAMLINE_META = {
    'i06': META_STRING,
    'i06-1': I06_1_META_STRING,
    'i06-2': META_STRING,
    'i10': META_STRING,
    'i10-1': I10_1_META_STRING,
    'i16': I16_META_STRING,
    'i21': META_STRING,
}


"""
NXtransformations
code taken from https://github.com/DanPorter/i16_diffractometer
"""
import hdfmap
import numpy as np
import h5py

import mmg_toolbox.nexus.nexus_names as nn
from mmg_toolbox.utils.units import METERS
from mmg_toolbox.utils.rotations import norm_vector, rotation_t_matrix, translation_t_matrix, transform_by_t_matrix
from mmg_toolbox.nexus.nexus_functions import nx_find_all, bytes2str

H5pyType = h5py.File | h5py.Group | h5py.Dataset


def get_depends_on(root: None | str | H5pyType) -> str:
    """Return depends_on value from group or dataset"""
    if isinstance(root, h5py.Group):
        if nn.NX_DEPON in root:
            return str(root[nn.NX_DEPON].asstr()[...])
        else:
            raise Exception(f"group: {root} does not contain 'depends_on'")
    elif isinstance(root, h5py.Dataset):
        if nn.NX_DEPON in root.attrs:
            return bytes2str(root.attrs[nn.NX_DEPON])
        else:
            raise Exception(f"dataset: {root} does not contain 'depends_on'")
    elif not root:
        return '.'
    else:
        return root


def nx_depends_on_chain(path: str, hdf_file: h5py.Group) -> list[str]:
    """
    Returns list of paths in a transformation chain, linked by 'depends_on'
    :param path: hdf path of initial dataset or group
    :param hdf_file: Nexus file object
    :return:
    """
    if path in hdf_file:
        depends_on = get_depends_on(hdf_file[path])
    else:
        depends_on = path
    out = []
    if depends_on != '.':
        out.append(depends_on)
        out.extend(nx_depends_on_chain(depends_on, hdf_file))
    return out


def nx_direction(path: str, hdf_file: h5py.Group) -> np.ndarray:
    """
    Return a unit-vector direction from a dataset
    :param path: hdf path of NXtransformation path or component group with 'depends_on'
    :param hdf_file: Nexus file object
    :return: unit-vector array
    """
    obj = hdf_file[path]
    depends_on = get_depends_on(obj)
    if depends_on == '.':
        dataset = obj
    elif isinstance(obj, h5py.Group) and depends_on in obj:
        dataset = obj[depends_on]
    elif depends_on in hdf_file:
        dataset = hdf_file[depends_on]
    else:
        raise Exception(f"{depends_on} not in {obj} or {hdf_file}")

    vector = np.asarray(dataset.attrs.get(nn.NX_VECTOR, (0, 0, 0)))
    return norm_vector(vector)


def nx_transformations_max_size(path: str, hdf_file: h5py.Group) -> int:
    """
    Return the maximum dataset size from a chain of transformations
    :param path: hdf dataset path of NX transformation, or group containing 'depends_on'
    :param hdf_file: Nexus file object
    :return: int : largest dataset.size
    """
    dataset = hdf_file[path]
    dataset_size = dataset.size if isinstance(dataset, h5py.Dataset) else 0
    depends_on = get_depends_on(dataset)
    if depends_on != '.':
        size = nx_transformations_max_size(depends_on, hdf_file)
        return size if size > dataset_size else dataset_size
    return dataset_size


def nx_transformations(path: str, index: int, hdf_file: h5py.Group, print_output=False) -> list[np.ndarray]:
    """
    Create list of 4x4 transformation matrices matching transformations along an NXtransformations chain
    :param path: str hdf path of the first point in the chain (Group or Dataset)
    :param index: int index of point in scan
    :param hdf_file: Nexus file object
    :param print_output: bool, if true the operations will be printed
    :return: list of 4x4 arrays [T1, T2, T3, ... Tn]
    """
    dataset = hdf_file[path]
    depends_on = get_depends_on(dataset)
    if print_output:
        print(f"{dataset}, depends on: {depends_on}")

    if isinstance(dataset, h5py.Group):
        return nx_transformations(depends_on, index, hdf_file, print_output)

    this_index = index if dataset.size > 1 else 0
    value = dataset[np.unravel_index(this_index, dataset.shape)]

    transformation_type = dataset.attrs.get(nn.NX_TTYPE, b'').decode()
    vector = np.array(dataset.attrs.get(nn.NX_VECTOR, (1, 0, 0)))
    offset = dataset.attrs.get(nn.NX_OFFSET, (0, 0, 0))
    units = dataset.attrs.get(nn.NX_UNITS, b'').decode()

    if transformation_type == nn.NX_TROT:
        if print_output:
            print(f"Rotating about {vector} by {value} {units}  | {path}")
        if units == 'deg':
            value = np.deg2rad(value)
        elif units != 'rad':
            value = np.deg2rad(value)
            print(f"Warning: Incorrect rotation units: '{units}'")
        matrix = rotation_t_matrix(value, vector, offset)
    elif transformation_type == nn.NX_TTRAN:
        if print_output:
            print(f"Translating along {vector} by {value} {units}  | {path}")
        if units in METERS:
            unit_multiplier = METERS[units]
        else:
            unit_multiplier = 1.0
            print(f"Warning: unknown translation untis: {units}")
        value = value * unit_multiplier * 1000  # distance in mm
        matrix = translation_t_matrix(value, vector, offset)
    else:
        if print_output:
            print(f"transformation type of '{path}' not recognized: '{transformation_type}'")
        matrix = np.eye(4)

    if depends_on == '.':  # end chain
        return [matrix]
    return [matrix] + nx_transformations(depends_on, index, hdf_file, print_output)


def nx_transformations_matrix(path: str, index: int, hdf_file: h5py.Group) -> np.ndarray:
    """
    Combine chain of transformation operations into single matrix
    :param path: str hdf path of the first point in the chain (Group or Dataset)
    :param index: int index of point in scan
    :param hdf_file: Nexus file object
    :return: 4x4 array
    """
    matrices = nx_transformations(path, index, hdf_file)
    # Combine the transformations in reverse
    return np.linalg.multi_dot(matrices[::-1])  # multiply transformations Tn..T3.T2.T1


def nx_transform_vector(xyz, path: str, index: int, hdf_file: h5py.Group) -> np.ndarray:
    """
    Transform a vector or position [x, y, z] by an NXtransformations chain
    :param xyz: 3D coordinates, n*3 [[x, y, z], ...]
    :param path: hdf path of first object in NXtransformations chain
    :param index: int index of point in scan
    :param hdf_file: Nexus file object
    :return: n*3 array([[x, y, z], ...]) transformed by operations
    """
    xyz = np.reshape(xyz, (-1, 3))
    t_matrix = nx_transformations_matrix(path, index, hdf_file)
    return (np.dot(t_matrix[:3, :3], xyz.T) + t_matrix[:3, 3:]).T


########################################################################################################################
########################################## Transformation Classes ######################################################
########################################################################################################################


class NxTransformation:
    """
    Class containing single NXTransformation axis
    """
    def __init__(self, path: str, name: str, parent: str, value: float | np.ndarray, transformation_type: str,
                 vector: tuple[float, float, float], offset: tuple[float, float, float],
                 units: str, offset_units: str, depends_on: str):
        self.path = path
        self.name = name
        self.parent = parent
        self.value = value
        self.type = transformation_type
        self.vector = vector
        self.offset = offset
        self.units = units
        self.offset_units = offset_units
        self.depends_on = depends_on

    def __repr__(self):
        return f"NxTransformation('{self.path}', {self.type}||{self.vector}={self.value})"

    def __str__(self):
        if self.type == nn.NX_TTRAN:
            return f"Translating {self.parent} along {self.vector} by {self.value} {self.units}  | {self.path}"
        else:
            return f"Rotating {self.parent} about {self.vector} by {self.value} {self.units}  | {self.path}"

    def t_matrix(self):
        if self.type == nn.NX_TTRAN:
            return translation_t_matrix(self.value, self.vector, self.offset)
        else:
            return rotation_t_matrix(self.value, self.vector, self.offset)

    def transform(self, vec: np.ndarray) -> np.ndarray:
        return transform_by_t_matrix(vec, self.t_matrix())


def load_transformation(path: str, index: int, hdf_file: h5py.Group, parent: str = '') -> NxTransformation:
    """Read Transformation Operation from HDF file"""
    if nn.NX_VECTOR in hdf_file[path].attrs:
        dataset = hdf_file[path]
    else:
        depends_on = get_depends_on(hdf_file[path])
        if depends_on == '.':
            dataset = hdf_file[path]
        else:
            dataset = hdf_file[depends_on]

    this_index = index if dataset.size > 1 else 0
    value = dataset[np.unravel_index(this_index, dataset.shape)]

    transformation_type = dataset.attrs.get(nn.NX_TTYPE, b'').decode()
    vector = dataset.attrs.get(nn.NX_VECTOR, (1, 0, 0))
    offset = dataset.attrs.get(nn.NX_OFFSET, (0, 0, 0))
    units = dataset.attrs.get(nn.NX_UNITS, b'').decode()
    offset_units = dataset.attrs.get(nn.NX_OFFSET_UNITS, b'').decode()
    depends_on = dataset.attrs.get(nn.NX_DEPON, b'').decode()
    return NxTransformation(
        path=path,
        name=dataset.name.split('/')[-1],
        parent=parent,
        value=value,
        transformation_type=transformation_type,
        vector=vector,
        offset=offset,
        units=units,
        offset_units=offset_units,
        depends_on=depends_on
    )


class NxTransformationChain:
    """
    Class containing chain of transformation operations
    """
    def __init__(self, object: H5pyType, index: int = 0):
        if isinstance(object, h5py.Dataset) and nn.NX_TTYPE not in object.attrs:
            raise Exception(f"{object} does not have '{nn.NX_TTYPE}' attribute")
        elif isinstance(object, h5py.Group) and nn.NX_DEPON not in object:
            raise Exception(f"{object} does not contain {nn.NX_DEPON}")
        self.path = object.name
        self.name = self.path.split('/')[-1]
        self.index = index
        chain = nx_depends_on_chain(self.path, object.file)
        self.size = nx_transformations_max_size(self.path, object.file)
        self._chain = [
            load_transformation(_path, index, object.file, self.name) for _path in chain
        ]

    def __repr__(self):
        return f"NxTransformationChain('{self.path}', index={self.index+1}/{self.size})"

    def __str__(self):
        return repr(self) + '\n' + '\n'.join(str(t) for t in self._chain)

    def __getitem__(self, item):
        return self._chain[item]

    def __iter__(self):
        return iter(self._chain)

    def __len__(self):
        return len(self._chain)

    def t_matrix_list(self) -> list[np.ndarray]:
        return [t.t_matrix() for t in self._chain]

    def t_matrix(self) -> np.ndarray:
        if len(self) == 1:
            return self._chain[0].t_matrix()
        else:
            return np.linalg.multi_dot(self.t_matrix_list()[::-1])

    def transform(self, vec: np.ndarray) -> np.ndarray:
        return transform_by_t_matrix(vec, self.t_matrix())


def generate_nxtranformations_string(filename: str) -> str:
    """
    return a string describing all the transformation chains in the NeXus file
    """
    out_str = "######################## NXtransformations ##########################\n"
    with hdfmap.load_hdf(filename) as nxs:
        datasets = nx_find_all(nxs, nn.NX_DEPON)
        for dataset in datasets:
            chain = NxTransformationChain(dataset.parent, 0)
            out_str += str(chain) + '\n\n'
    return out_str


#TODO: Merge this and NxTransformation
class TransformationAxis:
    """Holder for data to define an NXtransformation dataset"""
    def __init__(self, name: str, value: float | np.ndarray,
                 transformation_type: str = 'rotation', units: str = 'Deg',
                 vector: tuple[float, float, float] = (1, 0, 0), offset: tuple[float, float, float] = (0, 0, 0),
                 offset_units: str = 'mm'):
        self.name = name
        self.value = value
        self.units = units
        self.type = transformation_type
        self.vector = vector
        self.offset = offset
        self.offset_units = offset_units


class RotationAxis(TransformationAxis):
    """Holder for data to define a rotation NXtransformation dataset with units Degrees"""
    def __init__(self, name: str, value: float | np.ndarray,
                 vector: tuple[float, float, float] = (1, 0, 0), offset: tuple[float, float, float] = (0, 0, 0),
                 offset_units: str = 'mm'):
        super().__init__(name, value, 'rotation', 'Deg', vector, offset, offset_units)


class TranslationAxis(TransformationAxis):
    """Holder for data to define a translation NXtransformation dataset with units mm"""
    def __init__(self, name: str, value: float | np.ndarray,
                 vector: tuple[float, float, float] = (1, 0, 0), offset: tuple[float, float, float] = (0, 0, 0),
                 offset_units: str = 'mm'):
        super().__init__(name, value, 'translation', 'mm', vector, offset, offset_units)



"""
Command line interface for Dataviewer
"""

import sys
import os

from mmg_toolbox.utils.env_functions import get_beamline, get_beamline_from_directory
from .misc.config import BEAMLINE_CONFIG, get_config
from .apps.experiment import create_title_window
from .apps.data_viewer import create_data_viewer
from .apps.nexus import create_nexus_viewer


def doc():
    from mmg_toolbox import tkguis
    help(tkguis)


def run(*args):
    """
    Command line interface for Dataviewer
    """
    if any(arg.lower() in ['-h', '--help', 'man'] for arg in args):
        doc()
        return

    beamline = next((bm for bm in BEAMLINE_CONFIG if bm in args), get_beamline())

    for n, arg in enumerate(args):
        if os.path.isdir(arg):
            beamline = get_beamline_from_directory(os.path.abspath(arg), beamline)
            config = get_config(beamline=beamline)
            create_data_viewer(arg, config=config)
            return
        elif os.path.isfile(arg):
            beamline = get_beamline_from_directory(os.path.abspath(arg), beamline)
            config = get_config(beamline=beamline)
            create_nexus_viewer(arg, config=config)
            return
    create_title_window(beamline)
    return


def cli_run():
    run(*sys.argv[1:])



"""
Functions to calculate rotation matrices.
"""

import numpy as np


def rotmatrix_z(phi):
    """
    Generate rotation matrix of phi Deg about z-axis (right handed)
    Equivalent to YAW in the Tain-Bryan convention
    Equivalent to -phi, -eta, -delta in You et al. diffractometer convention (left handed)
    :param phi: float angle in degrees
    :return: [3*3] array
    """
    phi = np.deg2rad(phi)
    c = np.cos(phi)
    s = np.sin(phi)
    return np.array([[c, -s, 0], [s, c, 0], [0, 0, 1]])


def rotmatrix_y(chi):
    """
    Generate rotation matrix of chi Deg about y-axis (right handed)
    Equivalent to PITCH in the Tain-Bryan convention
    Equivalent to chi in You et al. diffractometer convention (right handed)
    :param chi: float angle in degrees
    :return: [3*3] array
    """
    chi = np.deg2rad(chi)
    c = np.cos(chi)
    s = np.sin(chi)
    return np.array([[c, 0, s], [0, 1, 0], [-s, 0, c]])


def rotmatrix_x(mu):
    """
    Generate rotation matrix of mu Deg about x-axis (right handed)
    Equivalent to ROLL in the Tain-Bryan convention
    Equivalent to mu in You et al. diffractometer convention (right handed)
    :param mu: float angle in degrees
    :return: [3*3] array
    """
    mu = np.deg2rad(mu)
    c = np.cos(mu)
    s = np.sin(mu)
    return np.array([[1, 0, 0], [0, c, -s], [0, s, c]])


def rotmatrix_intrinsic(alpha, beta, gamma):
    """
    a rotation whose yaw, pitch, and roll angles are α, β and γ, respectively.
    More formally, it is an intrinsic rotation whose Tait–Bryan angles are α, β, γ, about axes z, y, x, respectively.
    https://en.wikipedia.org/wiki/Rotation_matrix#General_rotations
    :param alpha: float yaw angle in degrees
    :param beta: float pitch angle in degrees
    :param gamma: float gamma angle in degrees
    :return: [3*3] array
    """
    alpha = np.deg2rad(alpha)
    beta = np.deg2rad(beta)
    gamma = np.deg2rad(gamma)
    ca = np.cos(alpha)
    sa = np.sin(alpha)
    cb = np.cos(beta)
    sb = np.sin(beta)
    cg = np.cos(gamma)
    sg = np.sin(gamma)
    return np.array([[ca * cb, ca * sb * sg - sa * cg, ca * sb * cg + sa * sg],
                     [sa * cb, sa * sb * sg + ca * cg, sa * sb * cg - ca * sg], [-sb, cb * sg, cb * cg]])


def rotmatrix_diffractometer(phi, chi, eta, mu):
    """
    an intrinsic rotation using You et al. 4S+2D diffractometer
        mu right-handed rotation about x
        eta left-handed rotation about z'
        chi right-handed rotation about y''
        phi left-handed rotation about z'''
        Angles in degrees
      Z = MU.ETA.CHI.PHI
      V' = Z.V || rot_vec = np.dot(r, vec)
    :param phi: float left-handed rotation about z''' angle in degrees
    :param chi: float right-handed rotation about y'' angle in degrees
    :param eta: float left-handed rotation about z' angle in degrees
    :param mu: float right-handed rotation about x angle in degrees
    :return: [3*3] array
    """
    phi = np.deg2rad(phi)
    chi = np.deg2rad(chi)
    eta = np.deg2rad(eta)
    mu = np.deg2rad(mu)
    cp = np.cos(phi)
    sp = np.sin(phi)
    cc = np.cos(chi)
    sc = np.sin(chi)
    ce = np.cos(eta)
    se = np.sin(eta)
    cm = np.cos(mu)
    sm = np.sin(mu)
    r = np.array([
        [
            ce * cp * cc - se * sp,
            ce * sp * cc + se * cp,
            ce * sc
        ],
        [
            sm * cp * sc + cm * (-se * cp * cc - ce * sp),
            sm * sp * sc + cm * (ce * cp - se * sp * cc),
            -se * cm * sc - sm * cc
        ],
        [
            sm * (-se * cp * cc - ce * sp) - cm * cp * sc,
            sm * (ce * cp - se * sp * cc) - cm * sp * sc,
            cm * cc - se * sm * sc
        ]
    ])
    return r


def diffractometer(vec, phi, chi, eta, mu):
    """
    Perform an intrinsic rotation using You et al. 4S+2D diffractometer
        mu right-handed rotation about x
        eta left-handed rotation about z'
        chi right-handed rotation about y''
        phi left-handed rotation about z'''
    Z = MU.ETA.CHI.PHI
    Angles in degrees
    vec must be 1D or column vector (3*n)

    :param vec: [3*n] vector in the sample frame
    :param phi: float angle in degrees, left handed roation about z'''
    :param chi: float angle in degrees, right handed rotation about y''
    :param eta: float angle in degrees, left handed rotation about z'
    :param mu: float angle in degrees, right handed rotation about x
    :return:  [3*n] vector in the diffractometer frame
    """
    r = np.dot(rotmatrix_x(mu), np.dot(rotmatrix_z(-eta), np.dot(rotmatrix_y(chi), rotmatrix_z(-phi))))
    return np.dot(r, vec)


def detector_wavevector(delta, gamma, wavelength_a):
    """
    Calculate wavevector in diffractometer axis using detector angles
    :param delta: float angle in degrees in vertical direction (about diff-z)
    :param gamma: float angle in degrees in horizontal direction (about diff-x)
    :param wavelength_a: float wavelength in A
    :return: [1*3] wavevector position in A-1 == kf - ki
    """

    k = 2 * np.pi / wavelength_a
    delta = np.deg2rad(delta)
    gamma = np.deg2rad(gamma)
    sd = np.sin(delta)
    cd = np.cos(delta)
    sg = np.sin(gamma)
    cg = np.cos(gamma)
    return k * np.array([sd, cd * cg - 1, cd * sg])


def diffractometer2hkl(ub, phi=0, chi=0, eta=0, mu=0, delta=0, gamma=0, wavelength=1.0):
    """
    Return [h,k,l] position of diffractometer axes with given UB and wavelength
    :param ub: [3*3] array UB orientation matrix following Busing & Levy
    :param phi: float sample angle in degrees
    :param chi: float sample angle in degrees
    :param eta: float sample angle in degrees
    :param mu: float sample angle in degrees
    :param delta: float detector angle in degrees
    :param gamma: float detector angle in degrees
    :param wavelength: float wavelength in A
    :return: [h,k,l]
    """
    q = detector_wavevector(delta, gamma, wavelength)  # You Ql (12)
    z = np.dot(rotmatrix_x(mu), np.dot(rotmatrix_z(-eta), np.dot(rotmatrix_y(chi), rotmatrix_z(-phi))))  # You Z (13)

    inv_ub = np.linalg.inv(ub)
    inv_z = np.linalg.inv(z)

    hphi = np.dot(inv_z, q)
    return np.dot(inv_ub, hphi).T


def norm_vector(vector, min_mag=0.001):
    mag = np.linalg.norm(vector)
    if mag < min_mag:
        mag = 1.
    return np.divide(vector, mag)


def rot_matrix(angle_rad: float, axis=(0, 0, 1)):
    """
    Generate rotation matrix about arbitary axis
    https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle
    """
    axis = norm_vector(axis)
    ux, uy, uz = axis
    c = np.cos(angle_rad)
    s = np.sin(angle_rad)
    c1 = 1 - c
    r = np.array([
        [
            (ux * ux * c1) + c,
            (uy * ux * c1) - uz * s,
            (uz * ux * c1) + uy * s
        ],
        [
            (ux * uy * c1) + uz * s,
            (uy * uy * c1) + c,
            (uz * uy * c1) - ux * s,
        ],
        [
            (ux * uz * c1) - uy * s,
            (uy * uz * c1) + ux * s,
            (uz * uz * c1) + c,
        ]
    ])
    return r


def rotation_t_matrix(value=0.0, vector=(0, 0, 1), offset=(0, 0, 0)):
    """
    Create 4x4 transformation matrix including a rotation
    """
    t = np.eye(4)
    t[:3, :3] = rot_matrix(angle_rad=value, axis=vector)
    t[:3, 3] = offset
    return t


def translation_t_matrix(value=0.0, vector=(0, 0, 1), offset=(0, 0, 0)):
    """
    Create 4x4 transformation matrix including a translation
    """
    t = np.eye(4)
    translation = value * np.reshape(vector, 3) + np.reshape(offset, 3)
    t[:3, 3] = translation
    return t


def rotate_by_matrix(xyz, angle_deg=0.0, axis=(0, 0, 1)):
    r = rot_matrix(np.deg2rad(angle_deg), axis)
    xyz = np.reshape(xyz, (-1, 3))
    return np.dot(r, xyz.T).T


def transform_by_t_matrix(xyz, t_matrix):
    xyz = np.reshape(xyz, (-1, 3))
    return (np.dot(t_matrix[:3, :3], xyz.T) + t_matrix[:3, 3:]).T



"""
Miscellaneous functions
"""

import re
import numpy as np
from typing import Any
from collections import defaultdict


regex_number = re.compile(r'\d{3,}')
re_long_floats = re.compile(r'\d+\.\d{5,}')


def consolidate_strings(*args: tuple, match_symbol: str = '*', match_score: float = 0.6) -> list:
    """
    Consolidate list of strings by finding matching components and
    replace differences with a symbol
    :param args: list of strings
    :param match_symbol: str to replace non-matching
    :param match_score: float 0-1, difference score greater than this will be replaced
    :returns: reduced list of strings
    """
    new_list = set()  # create a set to remove duplicates
    matched_index = []
    for idx_a, a in enumerate(args[:-1]):
        if idx_a in matched_index:
            continue
        for idx_b, b in enumerate(args[idx_a+1:]):
            if len(a) == len(b):
                if sum(1 for n in range(len(a)) if a[n] == b[n]) / len(a) > match_score:
                    a = ''.join(a[n] if a[n] == b[n] else match_symbol for n in range(len(a)))
                    matched_index.append(idx_b + idx_a + 1)
        new_list.add(a)
    return list(new_list)


def consolidate_numeric_strings(*args: str) -> list:
    """
    Consolidate list of strings by finding numbers in strings
    :param args: list of strings
    :returns: reduced list of strings
    """
    strings = defaultdict(list)
    reset_strings = {}
    for arg in args:
        match = regex_number.search(arg)
        if match:
            number = int(match[0])
            new_str = regex_number.sub('####', arg)
            strings[new_str] += [number]
            reset_strings[new_str] = arg

    out = [
        f"{string} .. [{numbers2string(numbers)}]"
        if len(numbers) > 1 else reset_strings[string]
        for string, numbers in strings.items()
    ]
    return out


def findranges(scannos: list[int], sep=':') -> str:
    """
    Convert a list of numbers to a simple string
    E.G.
    findranges([1,2,3,4,5]) = '1:5'
    findranges([1,2,3,4,5,10,12,14,16]) = '1:5,10:2:16'
    """

    scannos = np.sort(scannos).astype(int)

    dif = np.diff(scannos)

    stt, stp, rng = [scannos[0]], [dif[0]], [1]
    for n in range(1, len(dif)):
        if scannos[n + 1] != scannos[n] + dif[n - 1]:
            stt += [scannos[n]]
            stp += [dif[n]]
            rng += [1]
        else:
            rng[-1] += 1
    stt += [scannos[-1]]
    rng += [1]

    out = []
    x = 0
    while x < len(stt):
        if rng[x] == 1:
            out += ['{}'.format(stt[x])]
            x += 1
        elif stp[x] == 1:
            out += ['{}{}{}'.format(stt[x], sep, stt[x + 1])]
            x += 2
        else:
            out += ['{}{}{}{}{}'.format(stt[x], sep, stp[x], sep, stt[x + 1])]
            x += 2
    return ','.join(out)


def numbers2string(scannos: list[int], sep=':') -> str:
    """
    Convert a list of numbers to a simple string
    E.G.
    numbers2string([50001,50002,50003]) = '5000[1:3]'
    numbers2string([51020,51030,51040]) = '510[20:10:40]'
    """

    if type(scannos) is str or type(scannos) is int or len(scannos) == 1:
        return str(scannos)

    scannos = np.sort(scannos).astype(str)

    n = len(scannos[0])
    while np.all([scannos[0][:-n] == x[:-n] for x in scannos]):
        n -= 1

    if n == len(scannos[0]):
        return '{}-{}'.format(scannos[0], scannos[-1])

    inistr = scannos[0][:-(n + 1)]
    strc = [i[-(n + 1):] for i in scannos]
    liststr = findranges(strc, sep=sep)
    return '{}[{}]'.format(inistr, liststr)


def round_string_floats(string: str) -> str:
    """
    Shorten string by removing long floats
    :param string: string, e.g. '#810002 scan eta 74.89533603616637 76.49533603616636 0.02 pil3_100k 1 roi2'
    :return: shorter string, e.g. '#810002 scan eta 74.895 76.495 0.02 pil3_100k 1 roi2'
    """
    def subfun(m):
        return str(round(float(m.group()), 3))
    return re_long_floats.sub(subfun, string)


def stfm(value: float, error: float) -> str:
    """
    Create standard form string from value and uncertainty"
     str = stfm(val,err)
     Examples:
          '35.25 (1)' = stfm(35.25,0.01)
          '110 (5)' = stfm(110.25,5)
          '0.0015300 (5)' = stfm(0.00153,0.0000005)
          '1.56(2)E+6' = stfm(1.5632e6,1.53e4)

    Notes:
     - Errors less than 0.01% of values will be given as 0
     - The maximum length of string is 13 characters
     - Errors greater than 10x the value will cause the value to be rounded to zero
    """

    # Determine the number of significant figures from the error
    if error == 0. or value / float(error) >= 1E5:
        # Zero error - give value to 4 sig. fig.
        out = '{:1.5G}'.format(value)
        if 'E' in out:
            out = '{}(0)E{}'.format(*out.split('E'))
        else:
            out = out + ' (0)'
        return out
    elif np.log10(np.abs(error)) > 0.:
        # Error > 0
        sigfig = np.ceil(np.log10(np.abs(error))) - 1
        dec = 0.
    elif np.isnan(error):
        # nan error
        return '{} (-)'.format(value)
    else:
        # error < 0
        sigfig = np.floor(np.log10(np.abs(error)) + 0.025)
        dec = -sigfig

    # Round value and error to the number of significant figures
    rval = round(value / (10. ** sigfig)) * (10. ** sigfig)
    rerr = round(error / (10. ** sigfig)) * (10. ** sigfig)
    # size of value and error
    pw = np.floor(np.log10(np.abs(rval if abs(rval) > 0 else 0.0001)))
    pwr = np.floor(np.log10(np.abs(rerr if abs(rerr) > 0 else 0.00001)))

    max_pw = max(pw, pwr)
    ln = max_pw - sigfig  # power difference

    if np.log10(np.abs(error)) < 0:
        rerr = error / (10. ** sigfig)

    # Small numbers - exponential notation
    if max_pw < -3.:
        rval = rval / (10. ** max_pw)
        fmt = '{' + '0:1.{:1.0f}f'.format(ln) + '}({1:1.0f})E{2:1.0f}'
        return fmt.format(rval, rerr, max_pw)

    # Large numbers - exponential notation
    if max_pw >= 4.:
        rval = rval / (10. ** max_pw)
        rerr = rerr / (10. ** sigfig)
        fmt = '{' + '0:1.{:1.0f}f'.format(ln) + '}({1:1.0f})E+{2:1.0f}'
        return fmt.format(rval, rerr, max_pw)

    fmt = '{' + '0:0.{:1.0f}f'.format(dec + 0) + '} ({1:1.0f})'
    return fmt.format(rval, rerr)


def shorten_string(string: str, max_length: int = 100, end_letters: int = 10) -> str:
    """
    Return a shortened version of the first line of the string

    e.g.
        s = '\n scan eta 74.89533603616637 76.49533603616636 0.02 pol hkl checkbeam msmapper euler pil3_100k 1 roi2'
        shorten_string(s) ->
        'scan eta 74.895 76.495 0.02 pil3_100k 1 roi2'

    Only the first non-empty line is returned.

    if after reducing floating point numbers the length of the string exceeds the maximum,
    the string is curtailed skipping upto the last end_letters of the first line of the string.

    :param string: string, e.g. '#810002 scan eta 74.895'
    :param max_length: maximum length of string
    :param end_letters: number of characters at the end of string
    :return: shortened string
    """
    string = next(ss for s in string.splitlines() if (ss := s.strip()))
    string = round_string_floats(string)
    if len(string) < max_length:
        return string
    end_string = string[-end_letters:] if end_letters > 0 else ''
    return string[:max_length - end_letters - 5] + ' ... ' + end_string


class DataHolder(dict):
    """
    Convert dict to object that looks like a class object with key names as attributes
    Replicates slightly the old scisoftpy.dictutils.DataHolder class, also known as DLS dat format.
        obj = DataHolder(**{'item1': 'value1'})
        obj['item1'] -> 'value1'
        obj.item1 -> 'value1'
        obj.keys() -> dict.keys()
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        for name in kwargs:
            setattr(self, name, kwargs[name])
            self.update({name: kwargs[name]})


def data_holder(scan_data: dict[str, np.ndarray], metadata: dict[str, Any]) -> DataHolder:
    """
    Create DataHolder object from scan data and metadata
    Return object that slightly replicates the old scisoftpy.dictutils.DataHolder class, also known as DLS dat format.
    """
    d = DataHolder(**scan_data)
    d.metadata = DataHolder(**metadata)
    return d




"""
Example Script
{{description}}
"""

import os
import matplotlib.pyplot as plt

from mmg_toolbox.xas.nxxas_loader import load_xas_scans
from mmg_toolbox.xas.spectra_container import average_polarised_scans


scan_files = [
    # {{filenames}}
    'file.nxs'
]
output_folder = '{{output_path}}'

# Load spectra from scans
scans = load_xas_scans(*scan_files)

# Plot raw spectra
fig, axes = plt.subplots(1, 2, figsize=(12, 6), dpi=80)
fig.suptitle('raw scan files')
for scan in scans:
    for n, (mode, spectra) in enumerate(scan.spectra.items()):
        spectra.plot(ax=axes[n], label=scan.name)
        axes[n].set_ylabel(mode)

for ax in axes.flat:
    ax.set_xlabel('E [eV]')
    ax.legend()

print('Normalise by pre-edge')
for scan in scans:
    scan.divide_by_preedge()
    print(scan)

# plot scan normalised scan files
fig, axes = plt.subplots(1, 2, figsize=(12, 6), dpi=80)
fig.suptitle('Normalise by pre-edge')
for scan in scans:
    for n, (mode, spectra) in enumerate(scan.spectra.items()):
        spectra.plot(ax=axes[n], label=scan.name)
        axes[n].set_ylabel(mode)

for ax in axes.flat:
    ax.set_xlabel('E [eV]')
    ax.legend()

print('Fit and subtract background')
for scan in scans:
    scan.auto_edge_background(peak_width_ev=3.)
    print(scan)
    new_filename = f"{os.path.splitext(scan.metadata.filename)[0]}_auto_edge_background.nxs"
    scan.write_nexus(os.path.join(output_folder, new_filename))

# Plot background subtracted scans
for scan in scans:
    fig, axes = plt.subplots(2, 2, figsize=(12, 10), dpi=80)
    fig.suptitle(scan.name)
    for n, (mode, spectra) in enumerate(scan.spectra.items()):
        spectra.plot_parents(ax=axes[0, n])
        spectra.plot_bkg(ax=axes[0, n])
        axes[0, n].set_ylabel(mode)

        spectra.plot(ax=axes[1, n], label=scan.name)
        axes[1, n].set_ylabel(mode)

    for ax in axes.flat:
        ax.set_xlabel('E [eV]')
        ax.legend()

print('\n\nAverage polarised scans')
# Scan polarisations
for scan in scans:
    print(f"{scan.name}: {scan.metadata.pol}")
pol1, pol2 = average_polarised_scans(*scans)
print(pol1)
print(pol2)

# Plot averaged scans
fig, axes = plt.subplots(1, 2, figsize=(12, 6), dpi=80)
fig.suptitle('Averaged polarised scans')
for scan in [pol1, pol2]:
    for n, (mode, spectra) in enumerate(scan.spectra.items()):
        spectra.plot(ax=axes[n], label=scan.name)
        axes[n].set_ylabel(mode)

for ax in axes.flat:
    ax.set_xlabel('E [eV]')
    ax.legend()


print('\n\nCalculate XMCD')
xmcd = pol1 - pol2
print(xmcd)

for name, spectra in xmcd.spectra.items():
    print(spectra)
    print(spectra.process)
    print(spectra.sum_rules_report(1))

# Save xmcd file
xmcd_filename = f"{scans[0].metadata.scan_no}-{scans[-1].metadata.scan_no}_{xmcd.name}.nxs"
xmcd.write_nexus(os.path.join(output_folder, xmcd_filename))

fig, axes = plt.subplots(1, 2, figsize=(12, 6), dpi=80)
fig.suptitle(xmcd.name.upper())
for n, (mode, spectra) in enumerate(xmcd.spectra.items()):
    spectra.plot(ax=axes[n])
    axes[n].set_ylabel(mode)

for ax in axes.flat:
    ax.set_xlabel('E [eV]')
    ax.legend()

plt.show()




from .functions import *
from .models import *
from .results import *
from .fit_functions import *
from .manager import *

__all__ = [
    'poisson_errors', 'peak_ratio', 'gen_weights', 'gauss', 'group_adjacent', 'find_peaks', 'find_peaks_str', 'max_index',
    'modelfit', 'peakfit', 'peak2dfit', 'generate_model', 'generate_model_script', 'multipeakfit',
    'peak_results', 'peak_results_str', 'peak_results_fit', 'peak_results_plot', 'Peak', 'FitResults',
    'PEAK_PARS', 'METHODS', 'PEAK_MODELS', 'BACKGROUND_MODELS', 'get_peak_model', 'get_background_model', 'get_default_model',
    'ScanFitManager'
]
